@@ -44,11 +44,14 @@ public class JavaBridge {
  static EventQueue accessQueue = null;
  static AccessibleContext oldFocusSource = null;
  static String oldFocusText = "";
+ static int oldFocusSelectStart = 0;
+ static int oldFocusSelectEnd = 0;
  static AccessibleContext oldFocusCtx = null;
  static AccessibleContext savedFocusSource = null;
  static java.lang.Object oldEventSource = null;
  static String oldEventName = null;
  static int oldDetail1 = -1;
+ static boolean debugFlag = false;
 
  javax.accessibility.Accessible appAccessible;
 
@@ -97,7 +100,8 @@ public class JavaBridge {
 				javax.accessibility.AccessibleRole role = null;
 				if (source instanceof AccessibleContext ) {
 					role = ((AccessibleContext) source).getAccessibleRole();
-					if ( role == javax.accessibility.AccessibleRole.TEXT ) 
+					if ( (role == javax.accessibility.AccessibleRole.TEXT) 
+					    || role.toDisplayString(java.util.Locale.US).equalsIgnoreCase("paragraph")) 
 					    textEvent = true;
 					else if ( role == javax.accessibility.AccessibleRole.MENU_BAR ) 
 					    dispatchFocusEvent (source);
@@ -114,18 +118,59 @@ public class JavaBridge {
 					    }
 					}
 				}
-//				else if (source instanceof AccessibleContext) 
-//					System.out.println ("Selection source role: " +
-//                                         ((AccessibleContext) source).getAccessibleRole());
-//				else if (source instanceof javax.accessibility.Accessible)
-//					System.out.println ("Selection role " +
-//				          ((javax.accessibility.Accessible) source).
-//							    getAccessibleContext().
-//							    getAccessibleRole());
 				if ( textEvent ) {
-				    dispatchEvent(source,
-						  "object:text-selection-changed", 
-						  0, 0);
+				    int selection_start = 0, selection_end = 0;
+				    int selection_change_start = 0, selection_change_end = 0;
+				    org.omg.CORBA.Any selection_any = AccessUtil.getORB().create_any ();
+				    AccessibleContext ac = null;
+				    if (source instanceof AccessibleContext) 
+					ac = (AccessibleContext) source;
+				    // if this is the currently focussed object, return the delta
+				    // otherwise return the whole thing
+				    AccessibleText text = ac.getAccessibleText ();
+				    boolean new_selection = true;
+				    if (text != null) {
+					selection_start = text.getSelectionStart ();
+				        selection_end = text.getSelectionEnd ();
+				    }
+				    if ((ac != null) && (ac == oldFocusSource)) {
+					if ((selection_start != oldFocusSelectStart) || 
+					    (selection_end != oldFocusSelectEnd)) {
+					    if (selection_start != oldFocusSelectStart) {
+						selection_change_start = selection_start;
+						if (oldFocusSelectEnd != oldFocusSelectStart) 
+						    selection_change_end = oldFocusSelectStart;
+						else
+						    selection_change_end = selection_end;
+					    }
+					    if (selection_end != oldFocusSelectEnd) {
+						selection_change_end = selection_end;
+						if (oldFocusSelectStart != oldFocusSelectEnd)
+						    selection_change_start = oldFocusSelectEnd;
+						else
+						    selection_change_start = selection_start;
+					    }
+					    oldFocusSelectStart = selection_start;
+					    oldFocusSelectEnd = selection_end;
+					}
+					else {
+					    new_selection = false;
+					}
+				    }
+				    else {
+					selection_change_start = selection_start;
+					selection_change_end = selection_end;
+				    }
+				    if (new_selection) {
+					selection_any.insert_string (oldFocusText.substring (
+									 Math.min (selection_change_start, 
+									      selection_change_end),
+									 Math.max (selection_change_start,
+									      selection_change_end)));
+					dispatchEvent(source,
+						      "object:text-selection-changed", 
+						      selection_start, selection_end, selection_any);
+				    }
 				}
 				else {
 				    dispatchEvent(source,
@@ -280,7 +325,12 @@ public class JavaBridge {
 	}
 
 	public JavaBridge() {
-		System.err.println ("Java Accessibility Bridge for GNOME loaded.\n");
+
+		setDebug();
+
+		if (debugFlag)
+			System.err.println ("Java Accessibility Bridge for GNOME loaded.\n");
+
 		// Not sure what kind of arguments should be sent to ORB
 		String vm_rev = System.getProperty("java.version");
 
@@ -373,7 +423,8 @@ public class JavaBridge {
 			accessible.ref();
 			AccessUtil.getRegistryObject().notifyEvent(event);
 		} catch (Exception ex) {
-			System.out.println ("Error dispatching event "+eventName+":"+eventSource+"; "+ex);
+			if (debugFlag)
+				System.out.println ("Error dispatching event "+eventName+":"+eventSource+"; "+ex);
 		}		
 	}
 
@@ -387,9 +438,17 @@ public class JavaBridge {
 	}
 
 	public static boolean dispatchDeviceEvent (DeviceEvent de) {
-		if (de == null) return false;
-		else return AccessUtil.getRegistryObject().
-			getDeviceEventController().notifyListenersSync(de);
+		try {
+			if (de == null) return false;
+			else return AccessUtil.getRegistryObject().
+				getDeviceEventController().notifyListenersSync(de);
+		}
+		catch (Exception e)
+		{
+			if (debugFlag)
+				System.out.println("Exception in dispatchDeviceEvent " + e);
+			return false;
+		}
 	}
 
 	public static void dispatchFocusEvent(java.lang.Object eventSource) {
@@ -453,6 +512,8 @@ public class JavaBridge {
 			    AccessibleText axt = ctx.getAccessibleText ();
 			    oldFocusText = getTextRange (axt, 0, 
 							 axt.getCharCount ());
+			    oldFocusSelectStart = axt.getSelectionStart ();
+			    oldFocusSelectEnd = axt.getSelectionEnd ();
 			}
 			else oldFocusText = "";
 		}
@@ -491,7 +552,8 @@ public class JavaBridge {
 			accessible.ref();
 			AccessUtil.getRegistryObject().notifyEvent(event);
 		} catch (Exception ex) {
-			System.out.println ("Error dispatching focus event "+":"+eventSource+":"+ex);
+			if (debugFlag)
+				System.out.println ("Error dispatching focus event "+":"+eventSource+":"+ex);
 		}		
 	}
 
@@ -500,6 +562,44 @@ public class JavaBridge {
 		int newCaretPosition = Integer.parseInt(e.getNewValue().toString());
 //        System.err.println("Carent EVENT : " + newCaretPosition + ", " + 
 //                           oldCaretPosition );
+		// check selection, to see if it has changed
+		AccessibleText text = null;
+		if ((e.getSource() instanceof AccessibleContext) && (e.getSource() == oldFocusSource)) {
+		    text = ((AccessibleContext )e.getSource()).getAccessibleText ();
+		    int selection_start = text.getSelectionStart ();
+		    int selection_end = text.getSelectionEnd ();
+		    if ((selection_start != oldFocusSelectStart) || 
+			(selection_end != oldFocusSelectEnd)) {
+			org.omg.CORBA.Any selection_any = AccessUtil.getORB().create_any ();
+			int selection_change_start = 0;
+			int selection_change_end = 0;
+			if (selection_start != oldFocusSelectStart) {
+			    selection_change_start = selection_start;
+			    if (oldFocusSelectEnd != oldFocusSelectStart)
+				selection_change_end = oldFocusSelectStart;
+			    else
+				selection_change_end = selection_end;
+			}
+			if (selection_end != oldFocusSelectEnd) {
+			    selection_change_end = selection_end;
+			    if (oldFocusSelectStart != oldFocusSelectEnd)
+				selection_change_start = oldFocusSelectEnd;
+			    else
+				selection_change_start = selection_start;
+			}
+			oldFocusSelectStart = selection_start;
+			oldFocusSelectEnd = selection_end;
+			if (selection_change_start != selection_change_end) {
+			    selection_any.insert_string (oldFocusText.substring (
+							     Math.min (selection_change_start, selection_change_end),
+							     Math.max (selection_change_start, selection_change_end)));
+			    dispatchEvent (e.getSource(),
+					   "object:text-selection-changed",
+					   selection_start, selection_end,
+					   selection_any);
+			}
+		    }
+		}
 		dispatchEvent(e.getSource(), 
 			      "object:text-caret-moved",
 			      newCaretPosition,
@@ -865,7 +965,8 @@ public class JavaBridge {
             }
             catch ( Exception e ) {
                 // System.err.println("AttachListener: exception caught " + e.toString() );
-                e.printStackTrace();
+		if (debugFlag) 
+			e.printStackTrace();
             }
         }
     }
@@ -954,23 +1055,31 @@ public class JavaBridge {
 				return false;
 			}
 
-		System.err.println ("Registering Application.");
-            appAccessible = new ApplicationAccessible ();
+			if (AccessUtil.getRegistryObject() == null)
+				return false;
+
+			if (debugFlag)
+				System.err.println ("Registering Application.");
+			appAccessible = new ApplicationAccessible ();
             
-            Application aref = getAccessibleObjectFactory().
+			Application aref = getAccessibleObjectFactory().
                                getApplication(appAccessible.getAccessibleContext());
             
-		    System.err.println(aref);
+			if (debugFlag)
+				System.err.println(aref);
             
-		    AccessUtil.getRegistryObject().registerApplication(aref);
+			AccessUtil.getRegistryObject().registerApplication(aref);
 		    
-		    System.err.println(aref.name() + ":" + aref.id());
-		    System.err.println("Just registered Application");
-		    return true;
+			if (debugFlag) {
+				System.err.println(aref.name() + ":" + aref.id());
+				System.err.println("Just registered Application");
+			}
+			return true;
 			
 		} catch (Exception e) {
-		    e.printStackTrace();
-		    return false;
+			if (debugFlag)
+				System.out.println("Error registering application, exception is: "+ e);
+			return false;
 		}
 	}
 	
@@ -1063,4 +1172,25 @@ public class JavaBridge {
 	  
   }
 
+	private void setDebug() {
+		try {
+			Process p = Runtime.getRuntime().exec ("env");
+			BufferedReader b = new BufferedReader (
+				new InputStreamReader (p.getInputStream ()));
+
+			String result = b.readLine();
+			if (result == null) 
+				return;
+
+			String str;
+			while ((str = b.readLine()) != null) {
+				if (str.indexOf ("GNOME_ACCESSIBILITY") >= 0) {
+					debugFlag = true;
+					return;
+				}
+			}
+		}
+		catch (Exception e) {
+		}
+	}
 }
